/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XPSignatures.h"


/////////////////////////////// Implementation ////////////////////////////////

#define SIGN_DATA_TO_SIGN_INDEX  0
#define SIGN_SIGNATURE_INDEX     1

BEGIN_IMPLEMENT_XP_PARAMETERS(CCryptoAPISignExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(2, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CCryptoAPISignExtendedStoredProcedure)

BOOL CCryptoAPISignExtendedStoredProcedure::CryptAcquireContext(HCRYPTPROV* phProv, LPCTSTR pszContainer, LPCTSTR pszProvider, DWORD dwProvType, DWORD dwFlags)
{
  //Try to access the CSP normally
  if (::CryptAcquireContext(phProv, pszContainer, pszProvider, dwProvType, dwFlags))
    return TRUE;

  //Ok, Try to acquire by creating a new key set
  if (!::CryptAcquireContext(phProv, pszContainer, pszProvider, dwProvType, CRYPT_MACHINE_KEYSET | CRYPT_NEWKEYSET | CRYPT_SILENT))
    return FALSE;

  //Try to generate a new signature key
  HCRYPTKEY hKey;
  if (!CryptGetUserKey(*phProv, AT_SIGNATURE, &hKey))
  {
    if (GetLastError() == NTE_NO_KEY)
    {
      if (!CryptGenKey(*phProv, AT_SIGNATURE, CRYPT_CREATE_SALT, &hKey))
      {
        //Release the CSP before we return
        CryptReleaseContext(*phProv, 0);

        return FALSE;
      }
      else
        CryptDestroyKey(hKey);
    }
    else
    {
      //Release the CSP before we return
      CryptReleaseContext(*phProv, 0);

      return FALSE;
    }
  }
  else
    CryptDestroyKey(hKey);

  //Try to generate a new key exchange key
  if (!CryptGetUserKey(*phProv, AT_KEYEXCHANGE, &hKey))
  {
    if (GetLastError() == NTE_NO_KEY)
    {
      if (!CryptGenKey(*phProv, AT_KEYEXCHANGE, CRYPT_CREATE_SALT, &hKey))
      {
        //Release the CSP before we return
        CryptReleaseContext(*phProv, 0);

        return FALSE;
      }
      else
        CryptDestroyKey(hKey);
    }
    else
    {
      //Release the CSP before we return
      CryptReleaseContext(*phProv, 0);

      return FALSE;
    }
  }
  else
    CryptDestroyKey(hKey);

  return TRUE;
}

BOOL CCryptoAPISignExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CryptAcquireContext(&hProv, _T("XP_CRYPTOAPI"), NULL, m_dwProviderType, CRYPT_MACHINE_KEYSET | CRYPT_SILENT)) 
  {
    //Create the hash object
    HCRYPTHASH hHash;
    if (CryptCreateHash(hProv, m_AlgID, 0, 0, &hHash)) 
    {
      //Next thing to do is hash the data passed to us
      if (CryptHashData(hHash, m_pParameterData[SIGN_DATA_TO_SIGN_INDEX].m_pData, m_pParameterData[SIGN_DATA_TO_SIGN_INDEX].m_cbActualLen, 0))
      { 
        //Next we need to determine the size of the signature
        DWORD dwSignatureSize = 0;
        if (CryptSignHash(hHash, AT_SIGNATURE, NULL, 0, NULL, &dwSignatureSize))
        {
          //Allocate the data for the signature on the stack because it will always be relatively small
          BYTE* pbySignature = (BYTE*) _alloca(dwSignatureSize);
          if (CryptSignHash(hHash, AT_SIGNATURE, NULL, 0, pbySignature, &dwSignatureSize))
          {
            //Return the data in the output parameter directy
          	bSuccess = (ParamSetOutput(SIGN_SIGNATURE_INDEX+1, pbySignature, dwSignatureSize, FALSE) == SUCCEED);
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to get signature, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to get signature size, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to hash data, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the hash object
      CryptDestroyHash(hHash);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



CRSA_MD2_SignExtendedStoredProcedure::CRSA_MD2_SignExtendedStoredProcedure()
{
  m_AlgID = CALG_MD2;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_MD2_SignExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_MD2_SIGN @DataToSign @Signature");
}



CRSA_MD4_SignExtendedStoredProcedure::CRSA_MD4_SignExtendedStoredProcedure()
{
  m_AlgID = CALG_MD4;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_MD4_SignExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_MD4_SIGN @DataToSign @Signature");
}



CRSA_MD5_SignExtendedStoredProcedure::CRSA_MD5_SignExtendedStoredProcedure()
{
  m_AlgID = CALG_MD5;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_MD5_SignExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_MD5_SIGN @DataToSign @Signature");
}



CRSA_SHA1_SignExtendedStoredProcedure::CRSA_SHA1_SignExtendedStoredProcedure()
{
  m_AlgID = CALG_SHA1;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_SHA1_SignExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_SHA1_SIGN @DataToSign @Signature");
}



CDSA_SignExtendedStoredProcedure::CDSA_SignExtendedStoredProcedure()
{
  m_AlgID = CALG_SHA1;
  m_dwProviderType = PROV_DSS_DH;
}

void CDSA_SignExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DSA_SHA1_SIGN @DataToSign @Signature");
}



#define VERIFY_DATA_TO_VERIFY_INDEX  0
#define VERIFY_SIGNATURE_INDEX       1
#define VERIFY_PUBLICKEY_INDEX       2
#define VERIFY_RESULT_INDEX          3

BEGIN_IMPLEMENT_XP_PARAMETERS(CCryptoAPIVerifyExtendedStoredProcedure)
  XP_INPUT_PARAMETER (1, NULL, XP_ALLOW_ANY, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (2, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER (3, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE, XP_PT_EMPTY)
  XP_OUTPUT_PARAMETER(4, NULL, XP_ALLOW_BIT | XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CCryptoAPIVerifyExtendedStoredProcedure)

BOOL CCryptoAPIVerifyExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CCryptoAPISignExtendedStoredProcedure::CryptAcquireContext(&hProv, _T("XP_CRYPTOAPI"), NULL, m_dwProviderType, CRYPT_MACHINE_KEYSET | CRYPT_SILENT)) 
  {
    //Import the public key into the CSP
    HCRYPTKEY hPublicKey;
    if (CryptImportKey(hProv, m_pParameterData[VERIFY_PUBLICKEY_INDEX].m_pData, m_pParameterData[VERIFY_PUBLICKEY_INDEX].m_cbActualLen, 0, 0, &hPublicKey))
    {
      //Create the hash object
      HCRYPTHASH hHash;
      if (CryptCreateHash(hProv, m_AlgID, 0, 0, &hHash)) 
      {
        //Next thing to do is hash the data passed to us
        if (CryptHashData(hHash, m_pParameterData[VERIFY_DATA_TO_VERIFY_INDEX].m_pData, m_pParameterData[VERIFY_DATA_TO_VERIFY_INDEX].m_cbActualLen, 0))
        { 
          BOOL bVerified = CryptVerifySignature(hHash, m_pParameterData[VERIFY_SIGNATURE_INDEX].m_pData, m_pParameterData[VERIFY_SIGNATURE_INDEX].m_cbActualLen, hPublicKey, NULL, 0);

          //convert the bVerified into the output parameter data type
          int nConvertedDataSize = Convert(SRVINT4, &bVerified, sizeof(bVerified), m_pParameterData[VERIFY_RESULT_INDEX].m_bType, m_pParameterData[VERIFY_RESULT_INDEX].m_pData, m_pParameterData[VERIFY_RESULT_INDEX].m_cbMaxLen);
          if ((nConvertedDataSize > 0))
          {
            //Return the data in the output parameter
          	bSuccess = (ParamSetOutput(VERIFY_RESULT_INDEX+1, m_pParameterData[VERIFY_RESULT_INDEX].m_pData, nConvertedDataSize, FALSE) == SUCCEED);
          }
          else
          {
            sprintf(szErrorMsg, "%s: Failed to convert the verified value to the required output type, Error:%d", m_szFunctionName, GetLastError());
            SendErrorMsg(szErrorMsg);
          }
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to hash data, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }

        //Release the hash object
        CryptDestroyHash(hHash);
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to create hash, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Destroy the key now that we are finished with it
      CryptDestroyKey(hPublicKey);          
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to import the public key, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}





CRSA_MD2_VerifyExtendedStoredProcedure::CRSA_MD2_VerifyExtendedStoredProcedure()
{
  m_AlgID = CALG_MD2;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_MD2_VerifyExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_MD2_Verify @DataToVerify @Signature @PublicKeyData @Verified");
}



CRSA_MD4_VerifyExtendedStoredProcedure::CRSA_MD4_VerifyExtendedStoredProcedure()
{
  m_AlgID = CALG_MD4;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_MD4_VerifyExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_MD4_Verify @Signature @PublicKeyData @Verified");
}



CRSA_MD5_VerifyExtendedStoredProcedure::CRSA_MD5_VerifyExtendedStoredProcedure()
{
  m_AlgID = CALG_MD5;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_MD5_VerifyExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_MD5_Verify @DataToVerify @Signature @PublicKeyData @Verified");
}



CRSA_SHA1_VerifyExtendedStoredProcedure::CRSA_SHA1_VerifyExtendedStoredProcedure()
{
  m_AlgID = CALG_SHA1;
  m_dwProviderType = PROV_RSA_FULL;
}

void CRSA_SHA1_VerifyExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_SHA1_Verify @Signature @PublicKeyData @Verified");
}



CDSA_VerifyExtendedStoredProcedure::CDSA_VerifyExtendedStoredProcedure()
{
  m_AlgID = CALG_SHA1;
  m_dwProviderType = PROV_DSS_DH;
}

void CDSA_VerifyExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DSA_MD2_Verify @Signature @PublicKeyData @Verified");
}



#define EXPORTPUBLIC_KEY_INDEX  0

BEGIN_IMPLEMENT_XP_PARAMETERS(CCryptoAPIExportPublicKeyExtendedStoredProcedure)
  XP_OUTPUT_PARAMETER(1, NULL, XP_ALLOW_VARBINARY | XP_ALLOW_BIGVARBINARY | XP_ALLOW_BIGBINARY | XP_ALLOW_IMAGE | XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR, TRUE, FALSE)
END_IMPLEMENT_XP_PARAMETERS(CCryptoAPIExportPublicKeyExtendedStoredProcedure)

BOOL CCryptoAPIExportPublicKeyExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Local buffer we use for reporting messages back to ODS
  char szErrorMsg[2048];

  //What will be the return value
  BOOL bSuccess = FALSE;

  //First thing to do is acquire a CSP
  HCRYPTPROV hProv;
  if (CCryptoAPISignExtendedStoredProcedure::CryptAcquireContext(&hProv, _T("XP_CRYPTOAPI"), NULL, m_dwProviderType, CRYPT_MACHINE_KEYSET | CRYPT_SILENT)) 
  {
    //Get the handle to the signature public key
    HCRYPTKEY hKey;
    if (CryptGetUserKey(hProv, m_dwKeySpec, &hKey))
    {
      //Determine the size of the blob to export to
      DWORD dwBlobSize;
      if (CryptExportKey(hKey, 0, PUBLICKEYBLOB, 0, NULL, &dwBlobSize))
      {
        //Allocate the data for the signature public key on the stack because it will always be relatively small
        BYTE* pbyPublicKeyBlob = (BYTE*) _alloca(dwBlobSize);
        if (CryptExportKey(hKey, NULL, PUBLICKEYBLOB, 0, pbyPublicKeyBlob, &dwBlobSize))
        {
          //Return the data in the output parameter directy
          bSuccess = (ParamSetOutput(EXPORTPUBLIC_KEY_INDEX+1, pbyPublicKeyBlob, dwBlobSize, FALSE) == SUCCEED);
        }
        else
        {
          sprintf(szErrorMsg, "%s: Failed to get signature public key, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);
        }
      }
      else
      {
        sprintf(szErrorMsg, "%s: Failed to get signature public key size, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);
      }

      //Release the key now that we are finished with it
      CryptDestroyKey(hKey);
    }
    else
    {
      sprintf(szErrorMsg, "%s: Failed to get signature public key, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);
    }

    //Release the CSP
    CryptReleaseContext(hProv, 0);
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to acquire CSP, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);
  }

  return bSuccess;
}



CRSA_ExportSignExtendedStoredProcedure::CRSA_ExportSignExtendedStoredProcedure()
{
  m_dwProviderType = PROV_RSA_FULL;
  m_dwKeySpec = AT_SIGNATURE;
}

void CRSA_ExportSignExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_EXPORT_SIGNKEY @PublicKeyData");
}



CRSA_ExportEncryptExtendedStoredProcedure::CRSA_ExportEncryptExtendedStoredProcedure()
{
  m_dwProviderType = PROV_RSA_FULL;
  m_dwKeySpec = AT_KEYEXCHANGE;
}

void CRSA_ExportEncryptExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_RSA_EXPORT_ENCRYPTKEY @PublicKeyData");
}


CDSA_ExportExtendedStoredProcedure::CDSA_ExportExtendedStoredProcedure()
{
  m_dwProviderType = PROV_DSS_DH;
  m_dwKeySpec = AT_SIGNATURE;
}

void CDSA_ExportExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DSA_EXPORT @PublicKeyData");
}
