/////////////////////////////// Includes //////////////////////////////////////

#include "stdafx.h"
#include "XP_DISKLOGProc.h"



/////////////////////////////// Macros / Defines //////////////////////////////

#define XP_DISKLOG_MAX_MESSAGE_LENGTH  4096
#define XP_DISKLOG_MAX_FILENAME_LENGTH _MAX_PATH



/////////////////////////////// Implementation ////////////////////////////////

// A very simple class we use to perform automatic closure of Win32 Handles

class CAutoWin32Handle
{
public:
//Constructors / Destructors
  CAutoWin32Handle(HANDLE Handle)
  {
    m_Handle = Handle;
  }
  ~CAutoWin32Handle()
  {
    if (m_Handle != INVALID_HANDLE_VALUE)
    {
      CloseHandle(m_Handle);
      m_Handle = INVALID_HANDLE_VALUE;
    }
  }

//Methods
  operator HANDLE() const 
  { 
    return m_Handle; 
  };

protected:
//Member variables
  HANDLE m_Handle;
};



// A very simple class which wraps a critical section ( We could also have used 
// CCriticalSection if we were using MFC or CComAutoCriticalSection if we 
// were using ATL, rather than rolling our own implementation

class CCritSection
{
public:
//Constructors / Destructors
  CCritSection() 
  { 
    InitializeCriticalSection(&m_cs); 
  };
  ~CCritSection() 
  { 
    DeleteCriticalSection(&m_cs); 
  };

//Methods
  void Lock()   
  { 
    EnterCriticalSection(&m_cs); 
  };
  void Unlock() 
  { 
    LeaveCriticalSection(&m_cs); 
  };

protected:
//Member variable
  CRITICAL_SECTION m_cs;
};



// A class which performs locking and unlocking of a critical section based on scope
// Again if we were using MFC, we could just use the CSingleLock class, rather than
// rolling our own implementation

class CCritSectionLocker
{
public:
//Constructors / Destructors
  CCritSectionLocker(CCritSection* pCS)
  {
    XP_ASSERT(pCS);
    m_pCS = pCS;
    m_pCS->Lock();
  }
  ~CCritSectionLocker()
  {
    XP_ASSERT(m_pCS);
    m_pCS->Unlock();
  }

protected:
  CCritSection* m_pCS;
};


//The global crtical section we use to serialize access to our log files
CCritSection g_CriticalSection;



// The implementation of the actual XP_DISKLOG XP itself

#define LOG_FILENAME_INDEX      0
#define LOG_MESSAGE_INDEX       1
#define LOG_UNICODEFORMAT_INDEX 2
#define LOG_USEGMT_INDEX        3
#define LOG_LCID_INDEX          4

BEGIN_IMPLEMENT_XP_PARAMETERS(CDiskLogExtendedStoredProcedure)
  XP_INPUT_PARAMETER(1, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_TTEXT)
  XP_INPUT_PARAMETER(2, NULL, XP_ALLOW_BIGVARCHAR | XP_ALLOW_BIGCHAR | XP_ALLOW_TEXT | XP_ALLOW_VARCHAR | XP_ALLOW_NTEXT | XP_ALLOW_NVARCHAR, FALSE, FALSE, XP_PT_EMPTY)
  XP_INPUT_PARAMETER(3, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT,                                                   FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(4, NULL, XP_ALLOW_INT1 | XP_ALLOW_INT2 | XP_ALLOW_INT4 | XP_ALLOW_BIT,                                                   FALSE, TRUE,  XP_PT_INT4)
  XP_INPUT_PARAMETER(5, NULL, XP_ALLOW_INT4,                                                                                                  FALSE, TRUE,  XP_PT_INT4)
END_IMPLEMENT_XP_PARAMETERS(CDiskLogExtendedStoredProcedure)

void CDiskLogExtendedStoredProcedure::DisplayUsage()
{
  SendInfoMsg("USAGE: XP_DISKLOG @FileName @Message [@Unicode [@UseGMT [@LCID ] ] ] ");
}

BOOL CDiskLogExtendedStoredProcedure::AutoConvertParameters()
{
  //Validate our variables
  XP_ASSERT(m_pParameterDefinitions);
  XP_ASSERT(m_pParameterData);

  //If the Unicode parameter is present and is set, then specify that we want the message as a UNICODE string,
  //otherwise we want an ASCII string.
  BOOL bUnicodeOutput = FALSE;
  if (m_pParameterData[LOG_UNICODEFORMAT_INDEX].m_bType != SRVNULL)
    Convert(m_pParameterData[LOG_UNICODEFORMAT_INDEX].m_bType, m_pParameterData[LOG_UNICODEFORMAT_INDEX].m_pData, m_pParameterData[LOG_UNICODEFORMAT_INDEX].m_cbActualLen, SRVINT4, &bUnicodeOutput, sizeof(bUnicodeOutput));
  if (bUnicodeOutput)
    m_pParameterDefinitions[LOG_MESSAGE_INDEX].m_RuntimeType = XP_PT_UNICODETEXT;
  else
    m_pParameterDefinitions[LOG_MESSAGE_INDEX].m_RuntimeType = XP_PT_ASCIITEXT;
    
  //Let the base class do its thing
  return CExtendedStoredProcedure::AutoConvertParameters();
}


BOOL CDiskLogExtendedStoredProcedure::Run()
{
  //Validate our parmeters  
  XP_ASSERT(m_pParameterData);

  //Get the parameters
  TCHAR* pszFileName = XP_TTEXT(m_pParameterData[LOG_FILENAME_INDEX].m_Data);
  XP_ASSERT(pszFileName);

  //Should we treat the file as Unicode
  BOOL bUnicodeOutput = FALSE;
  if (m_pParameterData[LOG_UNICODEFORMAT_INDEX].m_bType != SRVNULL)
    bUnicodeOutput = (XP_INT4(m_pParameterData[LOG_UNICODEFORMAT_INDEX].m_Data) != 0);

  //Should we use GMT times
  BOOL bUseGMT = FALSE;
  if (m_pParameterData[LOG_USEGMT_INDEX].m_bType != SRVNULL)
    bUseGMT = (XP_INT4(m_pParameterData[LOG_USEGMT_INDEX].m_Data) != 0);

  //Has a LCID been specified
  LCID lcid = LOCALE_SYSTEM_DEFAULT;
  if (m_pParameterData[LOG_LCID_INDEX].m_bType != SRVNULL)
    lcid = XP_INT4(m_pParameterData[LOG_LCID_INDEX].m_Data);

  //The buffer we use for reporting any errors
  char szErrorMsg[2048];

  //Get the current time
  SYSTEMTIME st;
  if (bUseGMT)
    GetSystemTime(&st);
  else
    GetLocalTime(&st);

  //Flesh out the replacement strings for the filename
  TCHAR pszLogFile[XP_DISKLOG_MAX_FILENAME_LENGTH];
  if (!MapReplaceableStrings(lcid, st, pszFileName, pszLogFile, XP_DISKLOG_MAX_FILENAME_LENGTH))
  {
    sprintf(szErrorMsg, "%s: Failed to replace strings for the FileName parameter (Length is limited to %d characters!)", m_szFunctionName, XP_DISKLOG_MAX_FILENAME_LENGTH-1);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Flesh out the replacement strings for the message itself. This is a little
  //more complex because we need to handle a possible ASCII or UNICODE input string.
  //To keep it simple we allocate two stack arrays (one UNICODE and one ASCII) and
  //depending on the "bUnicodeOutput" flag we map to the correct one. Bear in mind that 
  //using a fixed size buffer of XP_DISKLOG_MAX_MESSAGE_LENGTH (defined as 4096) 
  //means that log messages are limited to 4095 characters in length. If a 
  //message ends up creating more data than can fit in this limit, then you will 
  //get an error returned from MapReplaceableStrings (And not a stack 
  //overwrite which would be very very bad!!!)
  wchar_t pswzLogMessage[XP_DISKLOG_MAX_MESSAGE_LENGTH];
  char psazLogMessage[XP_DISKLOG_MAX_MESSAGE_LENGTH];
  BOOL bSuccess = FALSE;
  if (bUnicodeOutput)
  {
    wchar_t* pszwInputMessage = XP_UNICODETEXT(m_pParameterData[LOG_MESSAGE_INDEX].m_Data);
    bSuccess = MapReplaceableStrings(lcid, st, pszwInputMessage, pswzLogMessage, XP_DISKLOG_MAX_MESSAGE_LENGTH);
  }
  else
  {
    char* pszaInputMessage = XP_ASCIITEXT(m_pParameterData[LOG_MESSAGE_INDEX].m_Data);
    bSuccess = MapReplaceableStrings(lcid, st, pszaInputMessage, psazLogMessage, XP_DISKLOG_MAX_MESSAGE_LENGTH);
  }
  if (!bSuccess)
  {
    sprintf(szErrorMsg, "%s: Failed to replace strings for the Message parameter (Length is limited to %d characters!)", m_szFunctionName, XP_DISKLOG_MAX_MESSAGE_LENGTH-1);
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  //Check to see if the file exists (this value is required later on)
  BOOL bFileExists = (GetFileAttributes(pszLogFile) != INVALID_FILE_ATTRIBUTES);

  //Serialize access to our file via the global critical section. By using a class we can use the standard scoping
  //rules of C++ to ensure that the critical section is unlocked at the correct time.
  CCritSectionLocker cslocket(&g_CriticalSection);

  //Open up the file to log to (If we were using MFC, we could have used the CFile class)
  HANDLE hFile = CreateFile(pszLogFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile != INVALID_HANDLE_VALUE)
  {
    //Wrap the raw file handle in a C++ object, this will ensure we do not leak the file handle
    //even if we prematurely exit this function without calling CloseHandle. It basically
    //allows us to concentrate on our application specific code. For a comprehensive class which
    //implements this, please refer to the "CW32Handle" class on the author's web site at
    //"http://www.naughter.com/w3mfc.html"
    CAutoWin32Handle File(hFile);

    //Write out the standard Unicode Little Endian BOM if required
    if (!bFileExists)
    {
      if (bUnicodeOutput)
      {
        char* pszBOM = "\xFF\xFE";
        DWORD dwWritten = 0;
        if (!WriteFile(File, pszBOM, 2, &dwWritten, NULL))
        {
          sprintf(szErrorMsg, "%s: Failed to write UNICODE BOM, Error:%d", m_szFunctionName, GetLastError());
          SendErrorMsg(szErrorMsg);

          return FALSE;
        }
      }
    }
    else
    {
      //Seek to the end of the file
      if (SetFilePointer(File, 0, NULL, FILE_END) == 0xFFFFFFFF)
      {
        sprintf(szErrorMsg, "%s: Failed to seek to the end of the file to log to, Error:%d", m_szFunctionName, GetLastError());
        SendErrorMsg(szErrorMsg);

        return FALSE;
      }
    }

    //Write out the message
    DWORD dwWritten = 0;
    if (bUnicodeOutput)
      bSuccess = WriteFile(hFile, pswzLogMessage, (DWORD) (wcslen(pswzLogMessage)*sizeof(wchar_t)), &dwWritten, NULL);
    else
      bSuccess = WriteFile(hFile, psazLogMessage, (DWORD) strlen(psazLogMessage), &dwWritten, NULL);
    if (!bSuccess)
    {
      sprintf(szErrorMsg, "%s: Failed to write the message to the log file, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }

    //While we're at it, write out a new line
    if (bUnicodeOutput)
    {
      wchar_t* pszEOL = L"\r\n";
      bSuccess = WriteFile(hFile, pszEOL, 4, &dwWritten, NULL);
    }
    else
    {
      char* pszEOL = "\r\n";
      bSuccess = WriteFile(hFile, pszEOL, 2, &dwWritten, NULL);
    }
    if (!bSuccess)
    {
      sprintf(szErrorMsg, "%s: Failed to write the end of line after the message to the log file, Error:%d", m_szFunctionName, GetLastError());
      SendErrorMsg(szErrorMsg);

      return FALSE;
    }
  }
  else
  {
    sprintf(szErrorMsg, "%s: Failed to open the file to log to, Error:%d", m_szFunctionName, GetLastError());
    SendErrorMsg(szErrorMsg);

    return FALSE;
  }

  return TRUE;
}

int CDiskLogExtendedStoredProcedure::DaysInMonth(int Month, BOOL IsLeap)
{
  static int gs_MonthLength[13] = 
  {
    0,  31, 28, 31, 30, 31, 30,
    31, 31, 30, 31, 30, 31
  };

  int r = gs_MonthLength[Month] + ((IsLeap && Month == 2) ? 1 : 0);
  return r;
}

BOOL CDiskLogExtendedStoredProcedure::IsLeap(int Year)
{
  if ((Year % 100) == 0)                
    return ((Year % 400) == 0) ? TRUE : FALSE;
  else
    return ((Year % 4) == 0) ? TRUE : FALSE;
}

int CDiskLogExtendedStoredProcedure::DaysSinceJan0(const SYSTEMTIME& st)
{
  BOOL bLeap = IsLeap(st.wYear);

  int nDays = 0;
  for (int i=1; i<st.wMonth; i++)
    nDays += DaysInMonth(i, bLeap);
  nDays += st.wDay;

  return nDays;
}

int CDiskLogExtendedStoredProcedure::GetWeekOfYear(const SYSTEMTIME& st)
{
  return DaysSinceJan0(st) / 7;
}



BOOL CDiskLogExtendedStoredProcedure::GetAbrStringDayOfWeek(LCID lcid, const SYSTEMTIME& st, LPSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wDayOfWeek)
  {
    case 1: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME1, pszBuf, nBufSize); break;
    case 2: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME2, pszBuf, nBufSize); break;
    case 3: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME3, pszBuf, nBufSize); break;
    case 4: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME4, pszBuf, nBufSize); break;
    case 5: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME5, pszBuf, nBufSize); break;
    case 6: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME6, pszBuf, nBufSize); break;
    case 0: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVDAYNAME7, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetFullStringDayOfWeek(LCID lcid, const SYSTEMTIME& st, LPSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wDayOfWeek)
  {
    case 1: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME1, pszBuf, nBufSize); break;
    case 2: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME2, pszBuf, nBufSize); break;
    case 3: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME3, pszBuf, nBufSize); break;
    case 4: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME4, pszBuf, nBufSize); break;
    case 5: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME5, pszBuf, nBufSize); break;
    case 6: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME6, pszBuf, nBufSize); break;
    case 0: nChars = GetLocaleInfoA(lcid, LOCALE_SDAYNAME7, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetAbrStringMonth(LCID lcid, const SYSTEMTIME& st, LPSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wMonth)
  {
    case 1:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME1, pszBuf, nBufSize);  break;
    case 2:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME2, pszBuf, nBufSize);  break;
    case 3:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME3, pszBuf, nBufSize);  break;
    case 4:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME4, pszBuf, nBufSize);  break;
    case 5:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME5, pszBuf, nBufSize);  break;
    case 6:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME6, pszBuf, nBufSize);  break;
    case 7:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME7, pszBuf, nBufSize);  break;
    case 8:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME8, pszBuf, nBufSize);  break;
    case 9:  nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME9, pszBuf, nBufSize);  break;
    case 10: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME10, pszBuf, nBufSize); break;
    case 11: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME11, pszBuf, nBufSize); break;
    case 12: nChars = GetLocaleInfoA(lcid, LOCALE_SABBREVMONTHNAME12, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetFullStringMonth(LCID lcid, const SYSTEMTIME& st, LPSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wMonth)
  {
    case 1:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME1,  pszBuf, nBufSize); break;
    case 2:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME2,  pszBuf, nBufSize); break;
    case 3:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME3,  pszBuf, nBufSize); break;
    case 4:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME4,  pszBuf, nBufSize); break;
    case 5:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME5,  pszBuf, nBufSize); break;
    case 6:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME6,  pszBuf, nBufSize); break;
    case 7:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME7,  pszBuf, nBufSize); break;
    case 8:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME8,  pszBuf, nBufSize); break;
    case 9:  nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME9,  pszBuf, nBufSize); break;
    case 10: nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME10, pszBuf, nBufSize); break;
    case 11: nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME11, pszBuf, nBufSize); break;
    case 12: nChars = GetLocaleInfoA(lcid, LOCALE_SMONTHNAME12, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetShortDateRepresentation(LCID lcid, const SYSTEMTIME& st, LPSTR pszBuf, int nBufSize)
{
  return (GetDateFormatA(lcid, DATE_SHORTDATE, &st, NULL, pszBuf, nBufSize) != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetLongDateRepresentation(LCID lcid, const SYSTEMTIME& st, LPSTR pszBuf, int nBufSize)
{
  return (GetDateFormatA(lcid, DATE_LONGDATE, &st, NULL, pszBuf, nBufSize) != 0);
}

//The function which expands the replaceable strings encounter in "psInputString" and returns the fleshed out result in pszOutputString. Note that
//the string in psInputString is not expected to be NULL terminated. Instead its string length is given by the parameter dwInputStringLength
BOOL CDiskLogExtendedStoredProcedure::MapReplaceableStrings(LCID lcid, const SYSTEMTIME& st, LPCSTR pszInputString, LPSTR pszOutputString, int nOutputStringLength)
{
  //Validate our parameters
  XP_ASSERT(pszInputString);
  XP_ASSERT(pszOutputString);
  XP_ASSERT(nOutputStringLength);

  int nInputStringLength = (int) strlen(pszInputString);

  //Start with a NULL terminated output string
  pszOutputString[0] = '\0';

  int nCurrentOutputStringLength = 0;
  for (int i=0; i<nInputStringLength; i++)
  {
    char c = pszInputString[i];
    if (c == '%')
    {
      ++i;
      if (i < nInputStringLength)
      {
        c = pszInputString[i];
        switch (c)
        {
          case 'h':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", st.wHour);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'i':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", st.wMinute);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 's':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", st.wSecond);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'z':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.03d", st.wMilliseconds);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'a':
          {
            char sBuffer[100];
            if (!GetAbrStringDayOfWeek(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'A':
          {
            char sBuffer[100];
            if (!GetFullStringDayOfWeek(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'b':
          {
            char sBuffer[100];
            if (!GetAbrStringMonth(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'B':
          {
            char sBuffer[100];
            if (!GetFullStringMonth(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'd':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", st.wDay);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'j':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.03d", DaysSinceJan0(st));
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'm':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", st.wMonth);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'U':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", GetWeekOfYear(st));
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'w':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%d", st.wDayOfWeek);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }                                   
          case 'x':
          {
            char sBuffer[100];
            if (!GetShortDateRepresentation(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }                 
          case 'y':
          { 
            char sBuffer[10];
            sprintf(sBuffer, "%.02d", st.wYear % 100);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case 'Y':
          {
            char sBuffer[10];
            sprintf(sBuffer, "%ld", st.wYear);
            nCurrentOutputStringLength += (int) strlen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            strcat(pszOutputString, sBuffer);
            break;
          }
          case '#':
          {
            if (i < nInputStringLength)
            {
              ++i;
              c = pszInputString[i];
              switch (c)
              {
                case 'h':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wHour);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 'i':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wMinute);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 's':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wSecond);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 'z':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wMilliseconds);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 'd':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wDay);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 'j':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", DaysSinceJan0(st));
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 'm':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wMonth);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                case 'U':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", GetWeekOfYear(st));
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }                                    
                case 'x':
                {
                  char sBuffer[100];
                  if (!GetLongDateRepresentation(lcid, st, sBuffer, 100))
                    return FALSE;
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }               
                case 'y':
                {
                  char sBuffer[10];
                  sprintf(sBuffer, "%d", st.wYear % 100);
                  nCurrentOutputStringLength += (int) strlen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  strcat(pszOutputString, sBuffer);
                  break;
                }
                default:
                {
                  ++nCurrentOutputStringLength;
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  char sBuffer[2];
                  sBuffer[0] = c;
                  sBuffer[1] = '\0';
                  strcat(pszOutputString, sBuffer);
                  break;
                }
              }
            }
            break;
          }
          default:
          {
            ++nCurrentOutputStringLength;
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            char sBuffer[2];
            sBuffer[0] = c;
            sBuffer[1] = '\0';
            strcat(pszOutputString, sBuffer);
            break;
          }
        }
      }
    }
    else
    {
      ++nCurrentOutputStringLength;
      if (nCurrentOutputStringLength >= nOutputStringLength)
        return FALSE;
      char sBuffer[2];
      sBuffer[0] = c;
      sBuffer[1] = '\0';
      strcat(pszOutputString, sBuffer);
    }
  }

  return TRUE;
}



BOOL CDiskLogExtendedStoredProcedure::GetAbrStringDayOfWeek(LCID lcid, const SYSTEMTIME& st, LPWSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wDayOfWeek)
  {
    case 1: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME1, pszBuf, nBufSize); break;
    case 2: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME2, pszBuf, nBufSize); break;
    case 3: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME3, pszBuf, nBufSize); break;
    case 4: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME4, pszBuf, nBufSize); break;
    case 5: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME5, pszBuf, nBufSize); break;
    case 6: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME6, pszBuf, nBufSize); break;
    case 0: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVDAYNAME7, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetFullStringDayOfWeek(LCID lcid, const SYSTEMTIME& st, LPWSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wDayOfWeek)
  {
    case 1: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME1, pszBuf, nBufSize); break;
    case 2: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME2, pszBuf, nBufSize); break;
    case 3: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME3, pszBuf, nBufSize); break;
    case 4: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME4, pszBuf, nBufSize); break;
    case 5: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME5, pszBuf, nBufSize); break;
    case 6: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME6, pszBuf, nBufSize); break;
    case 0: nChars = GetLocaleInfoW(lcid, LOCALE_SDAYNAME7, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetAbrStringMonth(LCID lcid, const SYSTEMTIME& st, LPWSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wMonth)
  {
    case 1:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME1, pszBuf, nBufSize);  break;
    case 2:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME2, pszBuf, nBufSize);  break;
    case 3:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME3, pszBuf, nBufSize);  break;
    case 4:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME4, pszBuf, nBufSize);  break;
    case 5:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME5, pszBuf, nBufSize);  break;
    case 6:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME6, pszBuf, nBufSize);  break;
    case 7:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME7, pszBuf, nBufSize);  break;
    case 8:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME8, pszBuf, nBufSize);  break;
    case 9:  nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME9, pszBuf, nBufSize);  break;
    case 10: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME10, pszBuf, nBufSize); break;
    case 11: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME11, pszBuf, nBufSize); break;
    case 12: nChars = GetLocaleInfoW(lcid, LOCALE_SABBREVMONTHNAME12, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetFullStringMonth(LCID lcid, const SYSTEMTIME& st, LPWSTR pszBuf, int nBufSize)
{
  int nChars=0;
  switch (st.wMonth)
  {
    case 1:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME1,  pszBuf, nBufSize); break;
    case 2:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME2,  pszBuf, nBufSize); break;
    case 3:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME3,  pszBuf, nBufSize); break;
    case 4:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME4,  pszBuf, nBufSize); break;
    case 5:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME5,  pszBuf, nBufSize); break;
    case 6:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME6,  pszBuf, nBufSize); break;
    case 7:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME7,  pszBuf, nBufSize); break;
    case 8:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME8,  pszBuf, nBufSize); break;
    case 9:  nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME9,  pszBuf, nBufSize); break;
    case 10: nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME10, pszBuf, nBufSize); break;
    case 11: nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME11, pszBuf, nBufSize); break;
    case 12: nChars = GetLocaleInfoW(lcid, LOCALE_SMONTHNAME12, pszBuf, nBufSize); break;
    default: XP_ASSERT(FALSE); break;
  }

  return (nChars != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetShortDateRepresentation(LCID lcid, const SYSTEMTIME& st, LPWSTR pszBuf, int nBufSize)
{
  return (GetDateFormatW(lcid, DATE_SHORTDATE, &st, NULL, pszBuf, nBufSize) != 0);
}

BOOL CDiskLogExtendedStoredProcedure::GetLongDateRepresentation(LCID lcid, const SYSTEMTIME& st, LPWSTR pszBuf, int nBufSize)
{
  return (GetDateFormatW(lcid, DATE_LONGDATE, &st, NULL, pszBuf, nBufSize) != 0);
}

//The function which expands the replaceable strings encounter in "psInputString" and returns the fleshed out result in pszOutputString. Note that
//the string in psInputString is not expected to be NULL terminated. Instead its string length is given by the parameter dwInputStringLength
BOOL CDiskLogExtendedStoredProcedure::MapReplaceableStrings(LCID lcid, const SYSTEMTIME& st, LPCWSTR pszInputString, LPWSTR pszOutputString, int nOutputStringLength)
{
  //Validate our parameters
  XP_ASSERT(pszInputString);
  XP_ASSERT(pszOutputString);
  XP_ASSERT(nOutputStringLength);

  int nInputStringLength = (int) wcslen(pszInputString);

  //Start with a NULL terminated output string
  pszOutputString[0] = '\0';

  int nCurrentOutputStringLength = 0;
  for (int i=0; i<nInputStringLength; i++)
  {
    wchar_t c = pszInputString[i];
    if (c == L'%')
    {
      ++i;
      if (i < nInputStringLength)
      {
        c = pszInputString[i];
        switch (c)
        {
          case L'h':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", st.wHour);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'i':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", st.wMinute);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L's':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", st.wSecond);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'z':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.03d", st.wMilliseconds);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'a':
          {
            wchar_t sBuffer[100];
            if (!GetAbrStringDayOfWeek(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'A':
          {
            wchar_t sBuffer[100];
            if (!GetFullStringDayOfWeek(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'b':
          {
            wchar_t sBuffer[100];
            if (!GetAbrStringMonth(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'B':
          {
            wchar_t sBuffer[100];
            if (!GetFullStringMonth(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'd':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", st.wDay);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'j':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.03d", DaysSinceJan0(st));
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'm':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", st.wMonth);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'U':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", GetWeekOfYear(st));
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'w':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%d", st.wDayOfWeek);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }                                   
          case L'x':
          {
            wchar_t sBuffer[100];
            if (!GetShortDateRepresentation(lcid, st, sBuffer, 100))
              return FALSE;
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }                 
          case L'y':
          { 
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%.02d", st.wYear % 100);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'Y':
          {
            wchar_t sBuffer[10];
            swprintf(sBuffer, L"%ld", st.wYear);
            nCurrentOutputStringLength += (int) wcslen(sBuffer);
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wcscat(pszOutputString, sBuffer);
            break;
          }
          case L'#':
          {
            if (i < nInputStringLength)
            {
              ++i;
              c = pszInputString[i];
              switch (c)
              {
                case L'h':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wHour);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L'i':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wMinute);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L's':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wSecond);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L'z':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wMilliseconds);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L'd':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wDay);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L'j':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", DaysSinceJan0(st));
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L'm':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wMonth);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                case L'U':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", GetWeekOfYear(st));
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }                                    
                case L'x':
                {
                  wchar_t sBuffer[100];
                  if (!GetLongDateRepresentation(lcid, st, sBuffer, 100))
                    return FALSE;
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }               
                case L'y':
                {
                  wchar_t sBuffer[10];
                  swprintf(sBuffer, L"%d", st.wYear % 100);
                  nCurrentOutputStringLength += (int) wcslen(sBuffer);
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
                default:
                {
                  ++nCurrentOutputStringLength;
                  if (nCurrentOutputStringLength >= nOutputStringLength)
                    return FALSE;
                  wchar_t sBuffer[2];
                  sBuffer[0] = c;
                  sBuffer[1] = L'\0';
                  wcscat(pszOutputString, sBuffer);
                  break;
                }
              }
            }
            break;
          }
          default:
          {
            ++nCurrentOutputStringLength;
            if (nCurrentOutputStringLength >= nOutputStringLength)
              return FALSE;
            wchar_t sBuffer[2];
            sBuffer[0] = c;
            sBuffer[1] = L'\0';
            wcscat(pszOutputString, sBuffer);
            break;
          }
        }
      }
    }
    else
    {
      ++nCurrentOutputStringLength;
      if (nCurrentOutputStringLength >= nOutputStringLength)
        return FALSE;
      wchar_t sBuffer[2];
      sBuffer[0] = c;
      sBuffer[1] = L'\0';
      wcscat(pszOutputString, sBuffer);
    }
  }

  return TRUE;
}
